def infer_columns(args, source_columns) -> dict:
    if 'end_pos' in args and args['end_pos']:
        source_columns[f"SUBSTRING_{args['target_col']}_{args['start_pos']}_{args['end_pos']}"] = 'varchar'
    else:
        source_columns[f"SUBSTRING_{args['target_col']}_{args['start_pos']}"] = 'varchar'
    return source_columns



"""
Functions to render Transform Templates
"""
from typing import Dict, Any, Optional
from pathlib import Path
from rasgotransforms.main import DataWarehouse
from rasgotransforms.dtypes import DTYPES
from importlib import machinery, util
import re


def get_dtype(dtype: str) -> str:
    return DTYPES[dtype.lower()]


def get_dw_type(dw_type: str) -> Optional[DataWarehouse]:
    if dw_type:
        try:
            return DataWarehouse[dw_type.upper()]
        except KeyError:
            raise Exception(f'Unsupported DataWarehouse type: {dw_type}')


def get_path(transform_name: str, dw_type: Optional[DataWarehouse] = None) -> Optional[str]:
    """
    Gets the python file for the given transform specific to the given data warehouse if one exists
    """
    root_dir = Path(__file__).parent.parent
    if dw_type:
        function_path = root_dir / 'transforms' / transform_name / dw_type.value / f'{transform_name}.py'
        if function_path.exists():
            return str(function_path.absolute())
    function_path = root_dir / 'transforms' / transform_name / f'{transform_name}.py'
    if function_path.exists():
        return str(function_path)


def cleanse_name(symbol: str) -> str:
    """
    Extra verbose function for clarity

    remove double quotes
    replace spaces and dashes with underscores
    cast to upper case
    delete anything that is not letters, numbers, or underscores
    if first character is a number, add an underscore to the beginning
    """
    symbol = str(symbol).strip().replace(' ', '_').replace('-', '_').upper()
    symbol = re.sub('[^A-Z0-9_]+', '', symbol)
    # if symbol is empty, at least returns '_'
    # if starts with a decimal prefix with '_'
    symbol = f'_{symbol}' if not symbol or symbol[0].isdecimal() else symbol

    return symbol


def infer_columns(
    transform_name: str, transform_args: Dict[str, Any], source_columns: Dict[str, str], dw_type: Optional[str] = None
) -> Optional[Dict[str, str]]:
    """
    Infers the column names and types in the output table from a query produced by a transform given
    the transform name, args, source columns, and data warehouse type
    """
    dw_type = get_dw_type(dw_type)
    function_path = get_path(transform_name, dw_type)
    if not function_path:
        return None
    cleaned_source_columns = {}
    for column_name, column_type in source_columns.items():
        cleaned_source_columns[cleanse_name(column_name)] = get_dtype(column_type)
    loader = machinery.SourceFileLoader(transform_name, function_path)
    spec = util.spec_from_loader(transform_name, loader)
    module = util.module_from_spec(spec)
    loader.exec_module(module)
    output_columns = module.infer_columns(args=transform_args, source_columns=source_columns)
    return {cleanse_name(name): get_dtype(column_type) for name, column_type in output_columns.items()}



from typing import Callable, Dict, Optional, Union, Tuple
from pathlib import Path
import inspect
import yaml

from rasgotransforms.render.environment import RasgoEnvironment
from rasgotransforms.main import DataWarehouse
from rasgotransforms.render.infer_columns import infer_columns
from rasgotransforms.exceptions import RenderException

__all__ = ["Transforms"]


class Transforms(object):
    """
    Dynamically adds class methods which render each transform in the package
    """

    def __init__(self, dw_type: str, run_query: Optional[Callable] = None, **kwargs):
        self.dw_type = DataWarehouse(dw_type.lower())
        self.run_query = run_query
        self.env = RasgoEnvironment(run_query=run_query, **kwargs)
        transforms_dir = Path(__file__).parent.parent / 'transforms'
        for transform_path in transforms_dir.rglob('*/*.yaml'):
            with open(transform_path) as fp:
                transform_config = yaml.safe_load(fp)
            setattr(self.__class__, transform_config['name'], self._get_render_method(transform_config))

    def _get_render_method(self, transform_config):
        """
        Creates the class method used to render jinja templates which is dynamically added to the transforms module
        """

        def render_transform(
            *args,
            source_table: str,
            source_columns: Optional[Dict[str, Dict[str, str]]] = None,
            return_columns: bool = False,
            **kwargs,
        ) -> Union[str, Tuple[str, Dict[str, str]]]:
            template_path = get_transform_path(name=transform_config['name'], dw_type=self.dw_type)
            with open(template_path) as fp:
                source_code = fp.read()
            sql = self.env.render(
                source_code=source_code, source_table=source_table, arguments=kwargs, source_columns=source_columns
            )
            if return_columns:
                if not source_columns:
                    raise RenderException('If return_columns is True, the source_columns must be passed in.')
                inferred_columns = infer_columns(
                    transform_name=transform_config['name'],
                    transform_args=kwargs,
                    source_columns=source_columns[source_table],
                    dw_type=self.dw_type.value,
                )
                return sql, inferred_columns
            else:
                return sql

        render_transform.__name__ = transform_config['name']
        render_transform.__signature__ = gen_method_signature(render_transform, transform_config)
        render_transform.__doc__ = gen_method_docstring(transform_config)
        return render_transform


def gen_method_signature(udt_func: Callable, transform_config: Dict) -> inspect.Signature:
    """
    Creates and returns a method signature.

    This is shown documentation for the parameters when hitting shift tab in a notebook
    """
    # Get current signature of function
    sig = inspect.signature(udt_func)

    params = []
    # Create Signature Params for Transform Args
    for arg in transform_config['arguments']:
        p = inspect.Parameter(name=arg, kind=inspect.Parameter.KEYWORD_ONLY)
        params.append(p)

    params += [
        inspect.Parameter(name='source_table', kind=inspect.Parameter.KEYWORD_ONLY),
        inspect.Parameter(name='source_columns', kind=inspect.Parameter.KEYWORD_ONLY),
    ]

    return sig.replace(parameters=params)


def gen_method_docstring(transform_config: Dict) -> str:
    """
    Generate and return a docstring for a transform method
    with transform description, args, and return specified.
    """
    # Have start of docstring be transform description
    docstring = f"\n{transform_config['description']}"

    # Add transform args to func docstring
    docstring = f"{docstring}\n  Args:"
    for arg, arg_meta in transform_config['arguments'].items():
        docstring = f"{docstring}\n    {arg}: {arg_meta['description']}"

    # Add return to docstring
    docstring = (
        f"{docstring}\n\n  Returns:\n    Returns an the rendered sql for the " f"{transform_config['name']} transform"
    )
    return docstring


def get_transform_path(name: str, dw_type: DataWarehouse):
    """
    Returns the default or data warehouse specific template file given the transform name
    """
    root_dir = Path(__file__).parent.parent
    if dw_type:
        function_path = root_dir / 'transforms' / name / dw_type.value / f'{name}.sql'
        if function_path.exists():
            return str(function_path.absolute())
    function_path = root_dir / 'transforms' / name / f'{name}.sql'
    if function_path.exists():
        return str(function_path)



def infer_columns(args, source_columns) -> dict:
    concat_list = [c.upper() for c in args['concat_list']]
    if 'overwrite_columns' in args and args['overwrite_columns']:
        out_cols = {k: v for k, v in source_columns.items() if k.upper() not in concat_list}
    else:
        out_cols = source_columns
    if 'alias' in args and args['alias']:
        out_cols[args['alias']] = 'text'
    else:
        out_cols[F'CONCAT_{"_".join(concat_list)}'] = 'text'
    return out_cols



def infer_columns(args, source_columns) -> dict:
    for col1 in args['columns1']:
        for col2 in args['columns2']:
            source_columns[f"{col1}_{col2}_Distance"] = 'int'
    return source_columns



def infer_columns(args, source_columns) -> dict:
    for target_col, date_part in args['dates'].items():
        source_columns[f"{target_col}_{date_part}".upper()] = source_columns[target_col.upper()]
    return source_columns



NUMERIC_TYPES = [
    'int',
    'integer',
    'bigint',
    'smallint',
    'number',
    'numeric',
    'float',
    'float4',
    'float8',
    'decimal',
    'double precision',
    'real',
]


def infer_columns(args, source_columns) -> dict:
    args = args.copy()
    out_cols = {}
    for col in args['group_by']:
        out_cols[col] = source_columns[col.upper()]
    if 'numeric columns' in args['aggregations'].keys():
        for column, column_type in source_columns.items():
            if column not in args['aggregations'].keys() and column_type.lower() in NUMERIC_TYPES:
                args['aggregations'].setdefault(column, []).extend(args['aggregations']['numeric columns'])
        args['aggregations'].pop('numeric columns')
    if 'nonnumeric columns' in args['aggregations'].keys():
        for column, column_type in source_columns.items():
            if column not in args['aggregations'].keys() and column_type.lower() not in NUMERIC_TYPES:
                args['aggregations'].setdefault(column, []).extend(args['aggregations']['nonnumeric columns'])
        args['aggregations'].pop('nonnumeric columns')
    for col in args['aggregations'].keys():
        for agg in args['aggregations'][col]:
            agg = agg.replace(' ', '')
            out_cols[f'{col}_{agg}'] = 'NUMERIC'
    return out_cols



def infer_columns(args, source_columns) -> dict:
    if 'flag_missing_vals' in args and args['flag_missing_vals']:
        for target_col, _ in args['replacements'].items():
            source_columns[f"{target_col}_missing_flag".upper()] = 'int'
    return source_columns



def infer_columns(args, source_columns) -> dict:
    source_columns[f"{args['date_col']}_spine_start"] = 'timestamp_ntz'
    source_columns[f"{args['date_col']}_spine_end"] = 'timestamp_ntz'
    return source_columns



"""
Functions to serve Rasgo Transform Templates
"""
import logging
import os
from enum import Enum
from collections import defaultdict
from pathlib import Path
from typing import Dict, List, Optional

import yaml

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class DataWarehouse(Enum):
    """
    Supported Data Warehouses
    """

    BIGQUERY = "bigquery"
    SNOWFLAKE = "snowflake"
    POSTGRESQL = "postgresql"
    MYSQL = "mysql"
    REDSHIFT = "redshift"


class TransformTemplate:
    """
    Reference to a Rasgo transform template
    """

    def __init__(
        self,
        name: str,
        arguments: List[dict],
        source_code: str,
        description: str = None,
        tags: List[str] = None,
        dw_type: Optional[str] = None,
    ):
        self.name = name
        self.arguments = arguments
        self.source_code = source_code
        self.description = description
        self.tags = tags
        self.dw_type = dw_type

    def __repr__(self) -> str:
        arg_str = ", ".join(f'{arg.get("name")}: {arg.get("type")}' for arg in self.arguments)
        return f"RasgoTemplate: {self.name}({arg_str})"

    def define(self) -> str:
        """
        Return a pretty string definition of this Transform
        """
        pretty_string = (
            f"Transform: {self.name}"
            f"\nDescription: {self.description}"
            f"\nArguments: {self.arguments}"
            f"\nSourceCode: {self.source_code}"
        )
        return pretty_string


def serve_rasgo_transform_templates(datawarehouse: str) -> List[TransformTemplate]:
    """
    Return a list of Rasgo Transform Templates

    Optionally pass in a datawarehouse name to include templates
    specific to its SQL dialect
    """
    datawarehouse = _check_datawarehouse(datawarehouse)

    template_list = []
    transform_yamls = _load_all_yaml_files()
    for transform_name, transform_yaml in transform_yamls.items():
        try:
            transform_source_code = _get_transform_source_code(
                transform_name=transform_name, datawarehouse=datawarehouse
            )
        except FileNotFoundError:
            # This allows for transforms for only specific data warehouses
            # otherwise raises without a base `sql` file in transform directory
            continue
        transform_args = _parse_transform_args_from_yaml(transform_yaml)
        template_list.append(
            TransformTemplate(
                name=transform_name,
                source_code=transform_source_code,
                arguments=transform_args,
                description=transform_yaml.get('description'),
                tags=transform_yaml.get('tags'),
            )
        )
    return template_list


def serve_rasgo_transform_snippets() -> List[Dict[str, str]]:
    """
    Returns a list of Rasgo Transform Snippets
    """
    output_list = []
    snippet_dir_path = _get_root_dir() / "snippets"
    snippets_files = [x for x in snippet_dir_path.rglob("*.sql")]
    for snippet in snippets_files:
        with open(snippet) as s_file:
            output_list.append({"snippet_name": snippet.name.replace(".sql", ""), "snippet_code": s_file.read()})
    return output_list


def _check_datawarehouse(input_value: str) -> str:
    supported_dws = "'" + "', '".join([e.value for e in DataWarehouse]) + "'"
    try:
        DataWarehouse[input_value.upper()]
    except Exception:
        raise ValueError(f'datawarehouse parameter accepts values: {supported_dws}')
    return input_value.lower()


def _get_root_dir() -> Path:
    """
    Get and return the root directory absolute path of this git repo
    """
    return Path(os.path.dirname(__file__))


def _get_transform_source_code(transform_name: str, datawarehouse: str) -> str:
    """
    Return a transform's source code as a string
    """
    datawarehouse = _check_datawarehouse(datawarehouse)
    root_dir = _get_root_dir()
    source_code_path = root_dir / "transforms" / transform_name / f"{transform_name}.sql"
    source_code_override_path = root_dir / "transforms" / transform_name / datawarehouse / f"{transform_name}.sql"
    if source_code_override_path.exists():
        source_code_path = source_code_override_path
    with open(source_code_path) as fp:
        source_code = fp.read()
    return source_code


def _load_all_yaml_files() -> Dict[str, Dict]:
    """
    Load and return all the yaml files in the dir <root>/<transform_type>_transforms
    """
    transform_yamls = defaultdict(dict)

    transform_dir_path = _get_root_dir() / "transforms"
    transform_names = [x.name for x in transform_dir_path.rglob("*/**")]
    for transform_name in transform_names:
        transform_yaml_path = transform_dir_path / transform_name / f"{transform_name}.yaml"

        try:
            transform_data = _read_yaml(transform_yaml_path)
            transform_yamls[transform_name] = transform_data
        except FileNotFoundError:
            continue

    return transform_yamls


def _parse_transform_args_from_yaml(transform_yaml: Dict) -> List[Dict[str, str]]:
    """
    From a loaded Transform Yaml File parse the
    Transform args in proper format, return the args
    in proper format for transform creation in PyRasgo
    """
    transform_args = []
    for arg_name, arg_meta_data in transform_yaml["arguments"].items():
        transform_args.append({**{"name": arg_name}, **arg_meta_data})
    return transform_args


def _read_yaml(yaml_path: Path) -> Dict:
    """
    Read and load a YAML file into a dictionary
    """
    with open(yaml_path, "r") as stream:
        try:
            return yaml.safe_load(stream)
        except yaml.YAMLError as e:
            logger.error(f"Error Parsing YAML file at {yaml_path}" f"\n\nError Msg: {e}")



def infer_columns(args, source_columns) -> dict:
    out_columns = {}
    for column_name, column_type in source_columns.items():
        if column_name not in args['group_by'] and column_name not in args['order_by']:
            out_columns[f"LATEST_{column_name}"] = column_type
        else:
            out_columns[column_name] = column_type
    return out_columns



def infer_columns(args, source_columns) -> dict:
    if 'alias' in args:
        source_columns[args['alias']] = 'int'
    else:
        source_columns[f"DIFF_{args['date_1']}_{args['date_2']}"] = 'int'
    return source_columns



def infer_columns(args, source_columns) -> dict:
    join_columns = args['source_columns'][args['join_table']]
    if 'join_prefix' in args and args['join_prefix']:
        for col_name, col_type in join_columns.items():
            source_columns[f"{args['join_prefix'].upper()}_{col_name}"] = col_type
        return source_columns
    else:
        return {**source_columns, **join_columns}



def infer_columns(args, source_columns) -> dict:
    if 'alias' in args and args['alias']:
        source_columns[args['alias']] = 'varchar'
    else:
        source_columns[f"REPLACE_{args['source_col']}"] = 'text'
    return source_columns



def infer_columns(args, source_columns) -> dict:
    for column in args['columns']:
        for amount in args['amounts']:
            source_columns[f"lag_{column}_{amount}".upper()] = source_columns[column.upper()]
    return source_columns



def infer_columns(args, source_columns) -> dict:
    for target_col, type in args['casts'].items():
        if 'overwrite_columns' in args and args['overwrite_columns']:
            source_columns[target_col.upper()] = type
        else:
            source_columns[f'{target_col.upper()}_{type.upper()}'] = type
    return source_columns



def infer_columns(args, source_columns) -> dict:
    out_columns = {}
    for column, column_type in source_columns.items():
        if column in args['renames']:
            out_columns[args['renames'][column]] = column_type
        else:
            out_columns[column] = column_type
    return out_columns



def infer_columns(args, source_columns) -> dict:
    for column in args['input_columns']:
        for window in args['window_sizes']:
            source_columns[f"mean_{column}_{window}"] = 'float'
    return source_columns



import re
from datetime import datetime, timedelta
from itertools import combinations, permutations, product
from typing import Callable, Optional, Dict
from pathlib import Path
from os.path import getmtime
from functools import partial

from jinja2 import Environment, BaseLoader
from jinja2.exceptions import TemplateNotFound
import json

from rasgotransforms.exceptions import RenderException
from rasgotransforms.main import DataWarehouse


class RasgoEnvironment(Environment):
    def __init__(self, run_query: Optional[Callable] = None, dw_type: Optional[str] = None, *args, **kwargs):
        super().__init__(*args, extensions=self.rasgo_extensions, loader=RasgoLoader(), **kwargs)
        if not dw_type:
            dw_type = 'snowflake'
        self._dw_type = DataWarehouse(dw_type)
        for filter_name, method in self.rasgo_filters.items():
            self.filters[filter_name] = method
        for name, value in self.rasgo_globals.items():
            self.globals[name] = value
        self._run_query = run_query
        self.globals['run_query'] = self._run_query

    @property
    def dw_type(self) -> DataWarehouse:
        return self._dw_type

    @dw_type.setter
    def dw_type(self, dw_type: str):
        self._dw_type = DataWarehouse(dw_type)

    @property
    def rasgo_extensions(self):
        return ['jinja2.ext.do', 'jinja2.ext.loopcontrols']

    @property
    def rasgo_globals(self):
        return {
            "adjust_start_date": partial(adjust_start_date, dw_type=self.dw_type),
            "cleanse_keys": cleanse_keys,
            "cleanse_name": cleanse_template_symbol,
            "combine_metrics": combine_metrics,
            "dw_type": lambda: self.dw_type.value,
            "get_timedelta": get_timedelta,
            "is_date_string": is_date_string,
            "itertools": {"combinations": combinations, "permutations": permutations, "product": product},
            "min": min,
            "max": max,
            "parse_comparison_value": partial(parse_comparison_value, dw_type=self.dw_type),
            "print": print,
            "quote": quote,
            "raise_exception": raise_exception,
            "ref_dataset": ref_not_found,
            "ref_metric": ref_not_found,
        }

    @property
    def rasgo_filters(self):
        return {
            "todatetime": datetime.fromisoformat,
            "to_json": json.dumps,
            "from_json": json.loads,
            "to_set": lambda x: set(x),
        }

    def render(
        self,
        source_code: str,
        arguments: dict,
        source_table: str = None,
        source_columns: Optional[Dict[str, Dict[str, str]]] = None,
        override_globals: Optional[Dict[str, Callable]] = None,
    ) -> str:
        """
        Given the source code of a jinja template, its arguments, and metadata about the source table,
        this renders the transform sql.

        Source columns can is a mapping of table names to columns (e.g. {table_name: {column_name: column_type}})
        """
        arguments['source_table'] = source_table

        if not override_globals:
            override_globals = {}

        if source_columns and 'get_columns' not in override_globals:

            def get_columns(fqtn):
                return source_columns[fqtn]

            override_globals['get_columns'] = get_columns
        if 'get_columns' in override_globals:
            self.globals['get_columns'] = override_globals['get_columns']
        try:
            template = self.from_string(source_code)
            rendered = template.render(**arguments, **override_globals)
        except Exception as e:
            raise RenderException(e)
        return trim_blank_lines(rendered)


def cleanse_keys(thing_with_keys):
    """
    If input object has dictionary keys, convert them from camel case to snake case
    """
    pattern = re.compile(r'(?<!^)(?=[A-Z])')

    def replace_key(key: str) -> str:
        return pattern.sub('_', key).lower()

    def cleanse_list_case(list_of_dicts: list) -> list:
        list_response = []
        for list_item in list_of_dicts:
            if isinstance(list_item, dict):
                list_response.append(cleanse_dict_case(list_item))
            elif isinstance(list_item, list):
                list_response.append(cleanse_list_case(list_item))
            else:
                list_response.append(list_item)
        return list_response

    def cleanse_dict_case(dictionary: dict) -> dict:
        dict_response = {}
        for k, v in dictionary.items():
            if isinstance(v, dict):
                dict_response[replace_key(k)] = cleanse_dict_case(v)
            elif isinstance(v, list):
                dict_response[replace_key(k)] = cleanse_list_case(v)
            else:
                dict_response[replace_key(k)] = v
        return dict_response

    if isinstance(thing_with_keys, dict):
        response = cleanse_dict_case(thing_with_keys)
    elif isinstance(thing_with_keys, list):
        response = cleanse_list_case(thing_with_keys)
    else:
        response = thing_with_keys
    return response


def cleanse_template_symbol(symbol: str) -> str:
    symbol = str(symbol).strip().replace(' ', '_').replace('-', '_')
    symbol = re.sub('[^A-Za-z0-9_]+', '', symbol)
    symbol = '_' + symbol if not symbol or symbol[0].isdecimal() else symbol
    return symbol


def combine_metrics(metrics: list) -> list:
    combined_metrics = {}
    output = []
    for metric in metrics:
        if metric.get('type', 'expression').lower() == 'expression':
            metric['names'] = [metric.get('name')]
            output.append(metric)
        else:
            combined_metrics.setdefault(metric.get('source_table'), []).append(metric)
    for source, metrics_to_combine in combined_metrics.items():
        combined_metric = metrics_to_combine[0]
        combined_metric['aggregations'] = []
        combined_metric['names'] = []
        for metric_to_combine in metrics_to_combine:
            combined_metric['names'].append(metric_to_combine.get('name'))
            combined_metric['aggregations'].append(
                {
                    'column': metric_to_combine.get('target_expression'),
                    'method': metric_to_combine.get('type'),
                    'alias': metric_to_combine.get('name'),
                }
            )
        output.append(combined_metric)
    return output


def raise_exception(message: str) -> None:
    raise Exception(message)


def ref_not_found(*args, **kwargs):
    raise NotImplementedError("`ref_` functions are only supported in the Rasgo API")


def trim_blank_lines(sql: str) -> str:
    return re.sub(r'[\n][\s]*\n', '\n', sql)


def get_timedelta(time_grain: str, interval: int = None) -> timedelta:
    time_grain = time_grain.lower()
    current_time = datetime.utcnow()
    if time_grain == 'hour':
        if not interval:
            return current_time - current_time.replace(minute=0, second=0, microsecond=0)
        return timedelta(hours=interval)
    elif time_grain == 'day':
        if not interval:
            return current_time - current_time.replace(minute=0, second=0, microsecond=0)
        return timedelta(days=interval)
    elif time_grain == 'week':
        if not interval:
            return timedelta(
                days=current_time.isoweekday(),
                hours=current_time.hour,
                minutes=current_time.minute,
                seconds=current_time.second,
                microseconds=current_time.microsecond,
            )
        return timedelta(weeks=interval)
    elif time_grain == 'month':
        if not interval:
            return current_time - current_time.replace(day=1, minute=0, second=0, microsecond=0)
        interval *= 31
        return timedelta(days=interval)
    elif time_grain == 'quarter':
        if not interval:
            return current_time - current_time.replace(
                month=((current_time.month - 1) // 3) * 3 + 1, day=1, minute=0, second=0, microsecond=0
            )
        interval *= 92
        return timedelta(days=interval)
    elif time_grain == 'year':
        if not interval:
            return current_time - current_time.replace(month=1, day=1, minute=0, second=0, microsecond=0)
        interval *= 365
        return timedelta(days=interval)
    else:
        raise RenderException(f"Invalid time grain '{time_grain}'")


def time_from_start_of_period(time_grain: str) -> timedelta:
    """
    Helper method that returns (current time - start of period/time grain)
    """
    time_grain = time_grain.lower()
    current_time = datetime.utcnow()
    if time_grain == 'hour':
        return current_time - current_time.replace(minute=0, second=0, microsecond=0)
    elif time_grain == 'day':
        return current_time - current_time.replace(minute=0, second=0, microsecond=0)
    elif time_grain == 'week':
        return timedelta(
            days=current_time.isoweekday(),
            hours=current_time.hour,
            minutes=current_time.minute,
            seconds=current_time.second,
            microseconds=current_time.microsecond,
        )
    elif time_grain == 'month':
        return current_time - current_time.replace(day=1, minute=0, second=0, microsecond=0)
    elif time_grain == 'quarter':
        return current_time - current_time.replace(
            month=((current_time.month - 1) // 3) * 3 + 1, day=1, minute=0, second=0, microsecond=0
        )
    elif time_grain == 'year':
        return current_time - current_time.replace(month=1, day=1, minute=0, second=0, microsecond=0)
    else:
        raise RenderException(f"Invalid time grain '{time_grain}'")


def parse_comparison_value(comparison_value, dw_type: DataWarehouse):
    if not isinstance(comparison_value, dict):
        return quote(comparison_value)
    direction = comparison_value.get('direction', 'none').lower()
    offset = comparison_value.get('offset', 0) or 0
    if direction == 'past':
        comparison_value['offset'] = -abs(offset)
    else:
        comparison_value['offset'] = abs(offset)
    date_part = comparison_value.get('date_part', comparison_value.get('datePart'))
    if date_part.lower() == 'all':
        date_part = 'DAY'
    if comparison_value['type'].lower() == 'relative_date':
        if dw_type == DataWarehouse.SNOWFLAKE:
            return f"DATEADD({date_part}, {comparison_value['offset']}, CURRENT_DATE)"
        elif dw_type == DataWarehouse.BIGQUERY:
            return f"DATE_ADD(CURRENT_DATE, INTERVAL {comparison_value['offset']} {date_part})"
        else:
            return f"(CURRENT_DATE + INTERVAL {comparison_value['offset']} {date_part})"
    elif comparison_value['type'].lower() == 'start_of_period':
        if dw_type == DataWarehouse.SNOWFLAKE:
            return f"DATE_TRUNC({date_part}, CURRENT_DATE)"
        elif dw_type == DataWarehouse.BIGQUERY:
            return f"DATE_TRUNC(CURRENT_DATE, {date_part})"
        else:
            return f"DATE_TRUNC({date_part}, CURRENT_DATE)"
    else:
        raise RenderException(f"Invalid comparison value object type '{comparison_value['type']}'")


def quote(value):
    """
    Helper method used to add quotes around strings only if they don't already exist
    """
    if isinstance(value, str) and not value.startswith("'"):
        return f"'{value}'"
    else:
        return value


def is_date_string(value: str) -> bool:
    """
    Test to determine if string is a date string
    """
    if not isinstance(value, str):
        return False
    try:
        datetime.fromisoformat(value)
        return True
    except ValueError:
        return False


def adjust_start_date(start_date, time_grain, secondary_calculations, dw_type: DataWarehouse):
    if not secondary_calculations:
        return parse_comparison_value(start_date, dw_type=dw_type)
    max_timedelta = timedelta(0)
    for calc in secondary_calculations:
        if 'interval' in calc:
            max_timedelta = max(max_timedelta, get_timedelta(time_grain, calc['interval']))
        elif 'period' in calc:
            max_timedelta = max(max_timedelta, get_timedelta(calc['period'], 1))

    if isinstance(start_date, dict):
        direction = start_date.get('direction', 'none').lower()
        offset = start_date.get('offset', 0)
        if direction == 'past':
            start_date['offset'] = -abs(offset)
        else:
            start_date['offset'] = abs(offset)
        date_part = start_date.get('date_part', start_date.get('datePart', None))
        if start_date['type'].lower() == 'start_of_period':
            time_delta = time_from_start_of_period(date_part)
        else:
            time_delta = get_timedelta(date_part, start_date['offset']) - max_timedelta
        return parse_comparison_value(
            {
                'type': 'relative_date',
                'offset': time_delta.days,
                'date_part': 'DAY',
                'direction': 'past' if time_delta.days < 0 else 'future',
            },
            dw_type=dw_type,
        )

    if isinstance(start_date, str):
        start_date = datetime.fromisoformat(start_date)
    return quote((start_date - max_timedelta).date().isoformat())


class RasgoLoader(BaseLoader):
    def __init__(self, root_path=None):
        if not root_path:
            root_path = Path(__file__).parent.parent / 'macros'
        self.root_path = root_path

    def get_source(self, environment: RasgoEnvironment, template):
        template_path = self.root_path / environment.dw_type.value / template
        if not template_path.exists():
            template_path = self.root_path / template
            if not template_path.exists():
                raise TemplateNotFound(template)
        mtime = getmtime(template_path)
        return template_path.read_text(), template_path, lambda: mtime == getmtime(template_path)



def infer_columns(args, source_columns) -> dict:
    if 'include_cols' in args:
        return {k: v for k, v in source_columns.items() if k in args['include_cols']}
    else:
        return {k: v for k, v in source_columns.items() if k not in args['exclude_cols']}



def infer_columns(args, source_columns) -> dict:
    if 'overwrite_columns' in args and args['overwrite_columns']:
        overwrite_columns = True
    else:
        overwrite_columns = False
    out_columns = source_columns.copy()
    for column_name, column_type in source_columns.items():
        if column_name in args['dates']:
            if overwrite_columns:
                out_columns[column_name] = 'date'
            else:
                out_columns[f"{column_name}_DATE"] = 'date'
    return source_columns



def infer_columns(args, source_columns) -> dict:
    if 'overwrite_columns' in args and args['overwrite_columns']:
        out_columns = {k: v for k, v in source_columns.items() if k not in args['rank_columns']}
    else:
        out_columns = source_columns.copy()
    if 'alias' in args and args['alias']:
        alias = args['alias']
    else:
        alias = f"RANK_{'_'.join(args['rank_columns'])}"
    out_columns[alias] = 'integer'
    return out_columns



DATE_PARTS = [
    'year',
    'month',
    'day',
    'dayofweek',
    'dayofweekiso',
    'dayofyear',
    'week',
    'weekiso',
    'quarter',
    'yearofweek',
    'yearofweekiso',
]
TIME_PARTS = [
    'hour',
    'minute',
    'second',
    'millisecond',
    'nanosecond',
    'epoch_second',
    'epoch_millisecond',
    'epoch_microsecond',
    'epoch_nanosecond',
    'timezone_hour',
    'timezone_minute',
]


def infer_columns(args, source_columns) -> dict:
    if args['date'] in source_columns:
        output_type = source_columns[args['date']]
    else:
        output_type = 'date'
    if 'overwrite_columns' in args and args['overwrite_columns']:
        source_columns[args['date'].upper()] = output_type
    elif 'alias' in args and args['alias']:
        source_columns[args['alias']] = output_type
    else:
        source_columns[f"{args['date']}_add{args['offset']}{args['date_part']}".upper()] = output_type
    return source_columns



import os
from setuptools import setup

_here = os.path.abspath(os.path.dirname(__file__))

version = {}
with open(os.path.join(_here, 'rasgotransforms', 'version.py')) as f:
    exec(f.read(), version)

with open(os.path.join(_here, 'DESCRIPTION.md'), encoding='utf-8') as f:
    long_description = f.read()

with open(os.path.join(_here, 'requirements.txt'), encoding='utf-8') as f:
    req_lines = f.read()
    requirements = req_lines.splitlines()

setup(
    name='rasgotransforms',
    version=version['__version__'],
    description=('Open-source SQL transform templates provided by RasgoML.'),
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Rasgo Intelligence',
    author_email='patrick@rasgoml.com',
    project_urls={
        'Documentation': 'https://docs.rasgoml.com/rasgo-docs/transforms/overview',
        'Source': 'https://github.com/rasgointelligence/RasgoTransforms',
        'Rasgo': 'https://www.rasgoml.com/',
    },
    license='GNU Affero General Public License v3 or later (AGPLv3+)',
    packages=[
        'rasgotransforms',
        'rasgotransforms/snippets',
        'rasgotransforms/transforms',
        'rasgotransforms/render',
    ],
    package_data={
        'rasgotransforms': [
            'transforms/**/*.yaml',
            'transforms/**/*.sql',
            'transforms/**/*.py',
            'transforms/**/**/*.sql',
            'transforms/**/**/*.py',
            'snippets/*.sql',
            'macros/*.sql',
            'macros/**/*.sql',
        ]
    },
    install_requires=requirements,
    include_package_data=True,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Database',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Software Development :: Code Generators',
    ],
    zip_safe=False,
)



def infer_columns(args, source_columns) -> dict:
    out_columns = {k: v for k, v in source_columns.items() if k in args['group_by']}
    out_columns['Slope'] = 'double'
    out_columns['Intercept'] = 'double'
    out_columns['R2'] = 'double'
    out_columns['Formula'] = 'varchar'
    return out_columns



DTYPES = {
    "smallint": "smallint",
    "bigint": "bigint",
    "int": "int",
    "integer": "integer",
    "tinyint": "tinyint",
    "byteint": "byteint",
    "float": "float",
    "float4": "float4",
    "float8": "float8",
    "float64": "float64",
    "decimal": "decimal",
    "numeric": "numeric",
    "number": "number",
    "real": "real",
    "double": "double",
    "string": "string",
    "text": "text",
    "varchar": "varchar",
    "char": "char",
    "character": "character",
    "date": "date",
    "datetime": "datetime",
    "time": "time",
    "timestamp": "timestamp",
    "timestamp_ltz": "timestamp_ltz",
    "timestamp_ntz": "timestamp_ntz",
    "timestamp_tz": "timestamp_tz",
    "binary": "binary",
    "varbinary": "varbinary",
    "boolean": "boolean",
    "bool": "bool",
    "variant": "variant",
    "object": "object",
    "array": "array",
    # Aliases
    "double precision": "double",
    "doubleprecision": "double",
}



def infer_columns(args, source_columns) -> dict:
    for column in args['columns_to_scale']:
        if not ('overwrite_columns' in args and args['overwrite_columns']):
            source_columns[f"{column}_MIN_MAX_SCALED"] = 'float'
        else:
            source_columns[column.upper()] = 'float'
    return source_columns



def infer_columns(args, source_columns) -> dict:
    if args['default'].upper() in source_columns:
        output_type = source_columns[args['default']]
    elif args['conditions'][1].upper() in source_columns:
        output_type = source_columns[args['conditions'][1].upper()]
    elif type(args['default']) is int:
        output_type = 'integer'
    elif type(args['default']) is float:
        output_type = 'float'
    else:
        output_type = 'text'
    source_columns[args['alias'].upper()] = output_type
    return source_columns
